# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from .pbori import (top_index, if_then_else,
                    substitute_variables, BooleSet,
                    ll_red_nf_redsb, ll_red_nf_noredsb,
                    ll_red_nf_noredsb_single_recursive_call)
from .PyPolyBoRi import (Polynomial, Monomial, Ring, BoolePolynomialVector)
from .statistics import used_vars_set
from .rank import rank

lead_index = top_index


def combine(reductors, p, reduce=None):
    p_nav = p.navigation()
    assert p_nav.value() < reductors.navigation().value()
    p_else = BooleSet(p_nav.else_branch(), p.ring())
    if reduce:
        p_else = reduce(p_else, reductors)
    return if_then_else(p_nav.value(), reductors, p_else)


def llredsb_Cudd_style(polys):

    reductors = Polynomial(polys[0].ring().one()).set() if polys else None

    linear_lead = sorted(polys, key=lead_index, reverse=True)
    assert len({p.lex_lead() for p in linear_lead}) == len(polys)
    assert not any(p.constant() for p in polys)
    assert len([p for p in polys if p.lex_lead_deg() == 1]) == len(polys)
    assert len({p.navigation().value() for p in polys}) == len(polys)
    for p in linear_lead:
        reductors = combine(reductors, p, reduce=ll_red_nf_redsb)
    return reductors


def ll_encode(polys, reduce=False, prot=False, reduce_by_linear=True):
    polys = [Polynomial(p) for p in polys]
    linear_lead = sorted(polys, key=lead_index, reverse=True)
    assert len({p.lex_lead() for p in linear_lead}) == len(polys)
    assert not any(p.constant() for p in polys)
    assert len([p for p in polys if p.lex_lead_deg() == 1]) == len(polys)
    assert len({p.navigation().value() for p in polys}) == len(polys)
    if (not reduce) and reduce_by_linear:
        linear_polys = [p for p in polys if p.deg() == 1]
        if linear_polys:
            linear_ll = ll_encode(linear_polys, reduce=True,
                                  reduce_by_linear=False)
            polys = [p.lex_lead() + ll_red_nf_redsb(p + p.lex_lead(),
                                                    linear_ll) for p in polys]
    reduce = ll_red_nf_redsb if reduce else None

    reductors = Polynomial(polys[0].ring().one()).set() if polys else None

    last = None
    counter = 0
    for p in linear_lead:

        if prot:
            counter = counter + 1
            progress = (counter * 100) / len(linear_lead)
            if last != progress:
                print(str(progress) + "%")
            last = progress
        reductors = combine(reductors, p, reduce=reduce)
    return reductors


def eliminate(polys, on_the_fly=False, prot=False, reduction_function=None,
              optimized=True):
    r"""
    There exists an optimized variant, which reorders the variable in a different ring.
    """
    polys = [Polynomial(p) for p in polys]
    rest = []
    linear_leads = []
    linear_leading_monomials = set()
    for p in polys:
        if p.is_zero():
            continue
        lm = p.lex_lead()
        if lm.deg() == 1:

            if lm not in linear_leading_monomials:
                linear_leading_monomials.add(lm)
                linear_leads.append(p)
            else:
                rest.append(p)
        else:
            rest.append(p)
    if not linear_leads:
        def identity(p):
            return p
        return (linear_leads, identity, rest)
    if reduction_function is None:
        if on_the_fly:
            if optimized:
                reduction_function = ll_red_nf_noredsb_single_recursive_call
            else:
                reduction_function = ll_red_nf_noredsb
        else:
            reduction_function = ll_red_nf_redsb

    if optimized:
        llnf, reduced_list = eliminate_ll_ranked(linear_leads, rest,
                                                 reduction_function=reduction_function,
                                                 reduce_ll_system=(not on_the_fly),
                                                 prot=prot)
    else:
        def llnf(p):
            return reduction_function(p, reductors)
        reduced_list = []
        reductors = ll_encode(linear_leads, reduce=(not on_the_fly), prot=prot)
        for p in rest:
            rp = reduction_function(p, reductors)
            if rp.is_one():
                reduced_list = [rp]
                break
            reduced_list.append(rp)

    return (linear_leads, llnf, reduced_list)


def construct_map_by_indices(to_ring, idx_mapping):
    v = BoolePolynomialVector((max(idx_mapping.keys()) + 1) * [to_ring.zero()])
    for (from_idx, to_idx) in idx_mapping.items():
        val = to_ring.variable(to_idx)
        v[from_idx] = val
    return v


def eliminate_ll_ranked(ll_system, to_reduce,
                        reduction_function=ll_red_nf_noredsb,
                        reduce_ll_system=False, prot=False):

    assert ll_system
    from_ring = ll_system[0].ring()

    ll_ranks = rank(ll_system)
    add_vars = set(used_vars_set(to_reduce).variables()).difference(ll_ranks.
                                                                    keys())
    for v in add_vars:
        ll_ranks[v] = -1

        # pushing variables ignored by ll to the front means,
        # that the routines will quickly eliminate them
        # and they won't give any overhead
    def sort_key(v):
        return (ll_ranks[v], v.index())
    sorted_vars = sorted(ll_ranks.keys(), key=sort_key)

    def var_index(v):
        return next(iter(Monomial(v).variables())).index()

    to_ring = Ring(len(sorted_vars))
    map_back_indices = {i: var_index(v) for i, v in enumerate(sorted_vars)}
    map_from_indices = {var_index(v): i for i, v in enumerate(sorted_vars)}

    var_names = [str(v) for v in sorted_vars]
    try:
        for (i, v) in enumerate(sorted_vars):
            assert var_names[i] == str(v), (var_names[i], v, var_index(v), i)

    finally:
        pass
    try:
        map_from_vec = construct_map_by_indices(to_ring, map_from_indices)
    finally:
        pass
    map_back_vec = construct_map_by_indices(from_ring, map_back_indices)

    def map_from(p):
        res = substitute_variables(to_ring, map_from_vec, p)
        return res

    def map_back(p):
        return substitute_variables(from_ring, map_back_vec, p)

    try:
        ll_opt_encoded = ll_encode([map_from(p) for p in ll_system],
                                   prot=False,
                                   reduce=reduce_ll_system)

        def llnf(p):
            return map_back(reduction_function(map_from(p), ll_opt_encoded))
        opt_eliminated = [llnf(p) for p in to_reduce]
    finally:
        pass
    return (llnf, opt_eliminated)


class RingMap:
    r"""
    Define a mapping between two rings by common variable names.

    TESTS::

        sage: from sage.rings.polynomial.pbori.pbori import *
        sage: from sage.rings.polynomial.pbori.blocks import declare_ring, Block
        sage: to_ring = declare_ring([Block("x", 10)], globals())
        sage: from_ring = declare_ring([Block("y", 5), Block("x", 10)], globals())
        sage: from sage.rings.polynomial.pbori.ll import RingMap
        sage: mapping = RingMap(to_ring, from_ring)
        sage: (x(1)+1).navigation().value()
        6
        sage: mapping(x(1)+1)
        x(1) + 1
        sage: mapping(x(1)+1).navigation().value()
        1
        sage: mapping.invert(mapping(x(1)+1))
        x(1) + 1
        sage: mapping.invert(mapping(x(1)+1)).navigation().value()
        6
        sage: mapping(y(1)+1)
        Traceback (most recent call last):
        ...
        RuntimeError: Operands come from different manager.
    """

    def __init__(self, to_ring, from_ring):
        r"""
        Initialize map by two given rings.

        TESTS::

            sage: from sage.rings.polynomial.pbori.pbori import *
            sage: from sage.rings.polynomial.pbori.blocks import declare_ring, Block
            sage: to_ring = declare_ring([Block("x", 10)], globals())
            sage: from_ring = declare_ring([Block("y", 5), Block("x", 10)], globals())
            sage: from sage.rings.polynomial.pbori.ll import RingMap
            sage: mapping = RingMap(to_ring, from_ring)
            sage: mapping(x(1)+1)
            x(1) + 1
        """
        def vars(ring):
            return [ring.variable(i) for i in range(ring.n_variables())]

        def indices(vars):
            return {str(v): idx for idx, v in enumerate(vars)}

        self.to_ring = to_ring
        self.from_ring = from_ring
        to_vars = vars(to_ring)
        from_vars = vars(from_ring)
        to_indices = indices(to_vars)
        from_indices = indices(from_vars)
        common = list(set(to_indices.keys()) & set(from_indices.keys()))

        to_map = list(from_vars)
        for elt in common:
            to_map[from_indices[elt]] = to_vars[to_indices[elt]]

        from_map = list(to_vars)
        for elt in common:
            from_map[to_indices[elt]] = from_vars[from_indices[elt]]

        self.to_map = BoolePolynomialVector(to_map)
        self.from_map = BoolePolynomialVector(from_map)

    def __call__(self, poly):
        r"""
        Execute the map to change rings.

        TESTS::

            sage: from sage.rings.polynomial.pbori.pbori import *
            sage: from sage.rings.polynomial.pbori.blocks import declare_ring, Block
            sage: to_ring = declare_ring([Block("x", 10)], globals())
            sage: from_ring = declare_ring([Block("y", 5), Block("x", 10)], globals())
            sage: from sage.rings.polynomial.pbori.ll import RingMap
            sage: mapping = RingMap(to_ring, from_ring)
            sage: mapping(x(1)+1)
            x(1) + 1
        """
        return substitute_variables(self.to_ring, self.to_map, poly)

    def invert(self, poly):
        r"""
        Inverted map to initial ring.

        EXAMPLES::

            sage: from sage.rings.polynomial.pbori.pbori import *
            sage: from sage.rings.polynomial.pbori.blocks import declare_ring, Block
            sage: to_ring = declare_ring([Block("x", 10)], globals())
            sage: from_ring = declare_ring([Block("y", 5), Block("x", 10)], globals())
            sage: from sage.rings.polynomial.pbori.ll import RingMap
            sage: mapping = RingMap(to_ring, from_ring)
            sage: mapping.invert(mapping(x(1)+1))
            x(1) + 1
        """
        return substitute_variables(self.from_ring, self.from_map, poly)



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
# Import basic functionality
r"""
This module defines an initial ring, and patches the declare_ring to use
a given context.

EXAMPLES::

    sage: from sage.rings.polynomial.pbori.frontend import x
    sage: x(0)
    x(0)
    sage: x(0)*x(0)
    x(0)
    sage: x(0) + x(0)
    0
    sage: x(9999)
    x(9999)
    sage: x(9999)*x(9999)
    x(9999)
    sage: x(9999) + x(9999)
    0

    sage: from sage.rings.polynomial.pbori.frontend import x, polybori_start
    sage: context = dict(globals())
    sage: polybori_start(context)
    ipbori...
    sage: r = context['declare_ring']('abc')
    sage: context['a']
    a
    sage: r.variable(0)
    a
"""


from .PyPolyBoRi import Ring
from .pbori import VariableFactory
from .blocks import declare_ring as orig_declare_ring


def block_scheme_names(blocks):
    r"""
    Helper for Singular interface.
    """
    context = {}
    from .blocks import declare_block_scheme
    declare_block_scheme(blocks, context)

    return list(context.keys())


ipbname = 'ipbori'


def polybori_start(global_context):
    def declare_ring(blocks, context=None):
        if context is None:
            context = global_context

        return orig_declare_ring(blocks, context)
    declare_ring.__doc__ = orig_declare_ring.__doc__
    global_context["declare_ring"] = declare_ring

    print(ipbname + """ -- The interactive command line tool of PolyBoRi/BRiAL %s
""" % global_context.get("polybori_version", ''))


# Here come the defaults
r = Ring(10000)
x = VariableFactory(r)



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
# Copyright (c) 2005-2007 by The PolyBoRi Team
from time import process_time as clock
from random import Random

from .PyPolyBoRi import (Polynomial, Variable, Monomial,
                         BoolePolynomialVector)
from .randompoly import gen_random_poly
from .pbori import (BooleSet, add_up_polynomials, interpolate_smallest_lex,
                    interpolate)
from .blocks import Block, declare_ring


generator = Random()


def add_up_poly_list(l, init):
    v = BoolePolynomialVector()
    for p in l:
        v.append(p)
    return add_up_polynomials(v, init)


def bench_interpolate(degree, nvariables, points):
    c = points
    h = len(points) / 2
    terms = set(c.terms())
    part1 = generator.sample(terms, h)
    part1 = add_up_poly_list(part1, Polynomial(c.ring().zero()))
    part2 = c + part1
    p = part1
    q = part2
    assert part1.set().intersect(part2).empty()
    c1 = clock()
    res2 = interpolate_smallest_lex(p, q)
    c2 = clock()
    print("finished interpolate_smallest_lex(p,q),len:", len(res2),
          "time", c2 - c1)
    c1 = clock()
    res1 = interpolate(p, q)
    c2 = clock()
    print("finished interpolate(p,q)" + len("_smallest_lex") * " " + ",len:",
          res1.set().size_double(), "time:", c2 - c1)
    return res2


def nf_lex_points(f, p):
    f = Polynomial(f)
    p = BooleSet(p)
    z = f.zeros_in(p)
    return interpolate_smallest_lex(z, p.diff(z))


def gen_random_o_z(points, points_p):
    k = generator.randrange(len(points) + 1)
    ones = generator.sample(points, k)
    vec = BoolePolynomialVector()
    for p in ones:
        vec.append(p)
    ones = add_up_polynomials(vec, Polynomial(points_p.ring().zero()))
    return interpolate_smallest_lex(points_p.set().diff(ones), ones)


def variety_lex_leading_terms(points, variables):
    assert isinstance(points, BooleSet), "Points needs to be a BooleSet"
    ring = variables.ring()
    standards = BooleSet(ring.zero())
    points_tuple = tuple(points)
    myvars_div = variables.divisors()
    if points != myvars_div:
        standards = BooleSet(ring.one())
    len_standards = len(standards)
    standards_old = standards
    while len_standards < len(points):
        standards = standards.union(gen_random_o_z(points_tuple, points))

        if standards_old != standards:
            standards = BooleSet(standards).include_divisors()
            len_standards = len(standards)
            standards_old = standards

    return BooleSet(myvars_div.diff(standards)).minimal_elements()


def lex_groebner_basis_points(points, variables):
    leads = variety_lex_leading_terms(points, variables)
    return [nf_lex_points(l, points) + l for l in leads]


def lex_groebner_basis_for_polynomial_via_variety(p):
    variables = p.vars_as_monomial()
    return lex_groebner_basis_points(p.zeros_in(variables.divisors()),
                                     variables)


if __name__ == '__main__':
    nvariables = 100
    r = declare_ring([Block("x", nvariables)])
    for number_of_points in (100, 500, 1000, 2000, 3000,
                             4000, 5000, 10000,
                             20000, 50000, 100000):
        print("----------")
        print("number_of_points:", number_of_points)
        print("generate points")
        points = gen_random_poly(r, number_of_points,
                                 nvariables,
                                 [Variable(i, r) for i in range(nvariables)])
        print("points generated")
        bench_interpolate(nvariables, nvariables, points)
        vars_mon = Monomial(r)
        for i in reversed(range(nvariables)):
            vars_mon = vars_mon * Variable(i, r)
        print(len(variety_lex_leading_terms(points, vars_mon)),
              "elements in groebner basis")



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
r"""
PolyBoRi's interface to libpolybori/BRiAL

This file makes interfaces to PolyBoRi's runtime libraries available in Python via sage.


AUTHOR:

- The PolyBoRi Team, 2007-2012

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori.pbori import *
        sage: from sage.rings.polynomial.pbori.blocks import declare_ring
        sage: r=declare_ring(["x0","x1","x2","y0","y1","y2"], globals())
        sage: x0>x1
        True
        sage: x0>x1*x2
        True
        sage: y0>y1
        True
        sage: y0>y1*y2
        True

        sage: r = r.clone(ordering=dlex)
        sage: r(x0) > r(x1)
        True
        sage: r(x0) > r(x1*x2)
        False

        sage: r = r.clone(ordering=dp_asc)
        sage: r(x0) > r(x1)
        False
        sage: r(x0) > r(x1*x2)
        False

        sage: r = r.clone(ordering=block_dlex, blocks=[3])
        sage: r(x0) > r(x1)
        True
        sage: r(x0) > r(x1*x2)
        False
        sage: r(x0) > r(y0*y1*y2)
        True

        sage: r = r.clone(ordering=block_dp_asc)
        sage: r(x0) > r(x1)
        False
        sage: r(x0) > r(y0)
        False
        sage: r(x0) > r(x1*x2)
        False

        sage: r = r.clone(ordering=block_dp_asc, blocks=[3])
        sage: r(x0) > r(y0)
        True

        sage: r(x0) > r(y0*y1)
        True

        sage: r = r.clone(names=["z17", "z7"])
        sage: [r.variable(idx) for idx in range(3)]
        [z17, z7, x2]
        sage: r = r.clone(names='abcde')
        sage: [r.variable(idx) for idx in range(6)]
        [a, b, c, d, e, y2]
"""

from .pbori import (order_dict, TermOrder_from_pb_order, BooleanPolynomialRing,
                    BooleanPolynomialVector, MonomialFactory,
                    PolynomialFactory, VariableFactory, add_up_polynomials)
from .pbori import gauss_on_polys as _gauss_on_polys

import weakref


OrderCode = type('OrderCode', (object,), order_dict)


def Ring(n, order='lp', names=None, blocks=None):
    if blocks is None:
        blocks = []
    pbnames = names
    if pbnames is None:
        pbnames = ['x(' + str(idx) + ')' for idx in range(n)]
    order = TermOrder_from_pb_order(n, order, blocks)
    return BooleanPolynomialRing(n, names=pbnames, order=order)


BoolePolynomialVector = BooleanPolynomialVector


# todo: PolyBoRi's original interface uses its WeakRingPtr here
def WeakRingRef(ring):
    return weakref.weakref(ring)


Monomial = MonomialFactory()
Polynomial = PolynomialFactory()
Variable = VariableFactory()


_add_up_polynomials = add_up_polynomials


def add_up_polynomials(polys, init):
    r"""
    Add up the polynomials in polys (which should be a
    ``BoolePolynomialVector`` or a sequence of ???
    """
    if not isinstance(polys, BoolePolynomialVector):
        vec = BoolePolynomialVector
        for p in polys:
            vec.append(p)
        polys = vec

    return _add_up_polynomials(polys, init)


def gauss_on_polys(l):
    vec = BoolePolynomialVector(l)
    return list(_gauss_on_polys(vec))



# sage_setup: distribution = sagemath-brial
from libcpp.memory cimport unique_ptr, shared_ptr, make_shared

from sage.rings.polynomial.multi_polynomial_ring_base cimport MPolynomialRing_base, BooleanPolynomialRing_base
from sage.rings.polynomial.multi_polynomial cimport MPolynomial
from sage.structure.element cimport MonoidElement

from sage.libs.polybori.decl cimport *


cdef class BooleanPolynomialRing(BooleanPolynomialRing_base):
    cdef PBRing _pbring
    cdef Py_ssize_t* pbind
    cdef public _monom_monoid
    cdef public object __interface
    cdef object _repr

    # it is very important to keep this cached, since otherwise the magma interface will break
    cdef public object __cover_ring

    cdef _convert(self, rhs)

cdef class BooleanPolynomial(MPolynomial):
    cdef PBPoly _pbpoly
    cpdef _add_(self, other)
    cpdef _mul_(self, other)

cdef class BooleSet:
    cdef BooleanPolynomialRing _ring
    cdef PBSet _pbset

cdef class CCuddNavigator:
    cdef PBNavigator _pbnav
    cdef Py_ssize_t* _pbind

cdef class BooleanMonomial(MonoidElement):
    cdef PBMonom _pbmonom
    cdef BooleanPolynomialRing _ring
    cpdef _mul_(self, other)

cdef class BooleanMonomialVariableIterator:
    cdef object parent
    cdef BooleanPolynomialRing _ring
    cdef BooleanMonomial obj
    cdef PBMonomIter _iter
    cdef PBMonomIter _end
    cdef Py_ssize_t* pbind

cdef class BooleanMonomialIterator:
    cdef BooleanMonomial obj
    cdef PBMonomIter _iter
    cdef PBMonomIter _end
    cdef Py_ssize_t* pbind

# Wrap PBPolyIter using pointers because there is no default constructor
cdef class BooleanPolynomialIterator:
    cdef BooleanPolynomial obj
    cdef PBPolyIter* _iter
    cdef PBPolyIter* _end

# Wrap PBSetIter using pointers because there is no default constructor
cdef class BooleSetIterator:
    cdef object _parent
    cdef BooleanPolynomialRing _ring
    cdef PBSetIter* _iter
    cdef PBSetIter* _end
    cdef BooleSet obj

cdef class BooleanPolynomialEntry:
    cdef public BooleanPolynomial p

# Wrap PBRedStrategy using shared_ptr because there is no default
# constructor and because multiple Python objects may point to
# the same PBRedStrategy
cdef class ReductionStrategy:
    cdef shared_ptr[PBRedStrategy] _strat
    cdef BooleanPolynomialRing _parent

# Wrap PBGBStrategy using shared_ptr because there is no default
# constructor and because multiple Python objects may point to
# the same PBGBStrategy
cdef class GroebnerStrategy:
    cdef shared_ptr[PBGBStrategy] _strat
    cdef BooleanPolynomialRing _parent
    cdef public ReductionStrategy reduction_strategy

# Wrap PBFGLMStrategy using unique_ptr for analogy with
# ReductionStrategy and GroebnerStrategy
cdef class FGLMStrategy:
    cdef unique_ptr[PBFGLMStrategy] _strat
    cdef BooleanPolynomialRing _parent

cdef class BooleanPolynomialVector:
    cdef PBPolyVector _vec
    cdef BooleanPolynomialRing _parent

cdef class BooleanPolynomialVectorIterator:
    cdef BooleanPolynomialVector obj
    cdef BooleanPolynomialRing _parent
    cdef PBPolyVectorIter _iter
    cdef PBPolyVectorIter _end

cdef class VariableBlock:
    cdef BooleanPolynomialRing _ring
    cdef PBVarBlock* _block
    cdef public object __name__

cdef class BooleConstant:
    cdef PBConstant _pbconst

cdef class VariableFactory:
    cdef BooleanPolynomialRing _ring
    cdef PBVarFactory _factory

cdef class MonomialFactory:
    cdef BooleanPolynomialRing _ring
    cdef PBMonomFactory _factory

cdef class PolynomialFactory:
    cdef BooleanPolynomialRing _ring
    cdef PBPolyFactory _factory


# Cython doesn't seem to support constructors with additional template
# parameters, so we declare this aliasing constructor as special case
cdef extern from *:
    cdef shared_ptr[T] shared_ptr_alias_PBGBStrategy "std::shared_ptr"[T](shared_ptr[PBGBStrategy]&, T*)



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
def input_signals(p):
    return list((p + p.lex_lead()).vars_as_monomial().variables())


def output_signal(p):
    return next(iter(p.lex_lead().variables()))


def rank(data):
    parents = {}
    res = {}
    for p in data:
        out = output_signal(p)
        parents.setdefault(out, [])
        for v in input_signals(p):
            parents.setdefault(v, []).append(out)

    def do_rank(v):
        if v in res:
            return res[v]
        my_res = res[v] = max((do_rank(p) + 1 for p in parents[v]), default=0)
        return my_res
    for v in parents:
        do_rank(v)
    return res



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
import sys
from itertools import chain, islice

from .pbori import VariableBlock
from .PyPolyBoRi import (Ring, Polynomial, VariableFactory, Variable)


class Block:
    r"""
    The block class represents a block of variables
    <var_name>(start_index,...,start_index+size-1), it is the preferred
    block type for simple one-dimensional variable sets
    """
    def __init__(self, var_name, size, start_index=0, reverse=False):
        indices = range(start_index, start_index + size)
        if reverse:
            indices = reversed(indices)
        self.names = [var_name + "(" + str(i) + ")" for i in indices]
        self.var_name = var_name
        self.start_index = start_index
        self.reverse = reverse
        self.size = size

    def __iter__(self):
        return iter(self.names)

    def __getitem__(self, i):
        return self.names[i]

    def __len__(self):
        return self.size

    def register(self, start, context):
        ring_context = context
        while isinstance(ring_context, PrefixedDictProxy):
            ring_context = ring_context.wrapped
        ring = ring_context['r']

        var_func = VariableBlock(self.size, self.start_index, start, self.
                                 reverse, ring)
        var_func.__name__ = self.var_name
        context[self.var_name] = var_func


class AlternatingBlock:
    r"""
    The Alternating Block class is used for doing tricky variable
    schemes,where base names vary, e.g.
    a(0),b(0),a(1),b(1),a(2),b(2)
    """
    def __init__(self, var_names, size_per_variable, start_index=0,
                 reverse=False):
        self.var_names = var_names
        self.size_per_variable = size_per_variable
        self.reverse = reverse
        indices = range(start_index, start_index + size_per_variable)

        if reverse:
            indices = reversed(indices)
        names = [f"{n}({str(i)})" for i in indices for n in var_names]
        self.indices = indices
        self.index2pos = {v: k for k, v in enumerate(indices)}
        self.names = names

    def __len__(self):
        return self.size_per_variable * len(self.var_names)

    def __iter__(self):
        return iter(self.names)

    def __getitem__(self, i):
        return self.names[i]

    def register(self, start, context):
        def gen_var_func(var_pos):

            class var_factory:
                def __init__(self, ring, index2pos, size):
                    self.ring = ring
                    self.index2pos = index2pos
                    self.size = size

                def __call__(self, idx):
                    return self.ring.variable(self.index2pos[idx] * self.size +
                                              var_pos + start)
            ring_context = context
            while isinstance(ring_context, PrefixedDictProxy):
                ring_context = ring_context.wrapped
            ring = ring_context['r']

            return var_factory(ring, self.index2pos, len(self.var_names))

        for (var_pos, n) in enumerate(self.var_names):
            var_func = gen_var_func(var_pos)
            var_func.__name__ = n
            context[n] = var_func


def shift(f, i):
    def g(j):
        return f(i + j)
    g.__name__ = f.__name__
    return g


class AdderBlock(AlternatingBlock):
    def __init__(self, adder_bits, sums='s', carries='c', input1='a',
                 input2='b', start_index=0):
        AlternatingBlock.__init__(self, (sums, carries, input1, input2),
                                  adder_bits, start_index=start_index,
                                  reverse=True)
        self.input1 = input1
        self.input2 = input2
        self.sums = sums
        self.carries = carries
        self.start_index = start_index
        self.adder_bits = adder_bits

    def register(self, start, context):
        super().register(start, context)
        a = context[self.input1]
        b = context[self.input2]
        self.s = shift(context[self.sums], self.start_index)
        self.c = shift(context[self.carries], self.start_index)
        a = shift(a, self.start_index)
        b = shift(b, self.start_index)
        carries = [Polynomial(a(0).ring().zero())]
        last = carries[0]
        for i in range(self.adder_bits):
            c = 1 + (1 + a(i) * b(i)) * (1 + last * a(i)) * (1 + last * b(i))
            carries.append(c)
            last = c

        self.add_results = [a(i) + b(i) + carries[i]
                            for i in range(self.adder_bits)]
        self.carries_polys = carries[1:]

    # def s(i):
    #     return self.add_results[i-self.start_index]
    # def c(i):
    #     return self.carries_polys[i-self.start_index]
    # context[self.sums]=s
    # context[self.carries]=c

    def implement(self, equations):
        for i in range(self.adder_bits):
            equations.append(self.s(i) + self.add_results[i])
            equations.append(self.c(i) + self.carries_polys[i])


class HigherOrderBlock:
    r"""
    HigherOrderBlocks are multidimensional blocks of variables.

    For each dimension a separate start_index and size can be specified.

    var_name : variables will be called <var_name>(multiindex), where
    multiindex is a tuple of the size <size_tuple>

    size_tuple : specifies the sizes of the ranges of each component of the
    multi-indices

    start_index_tuple : the multi-indices will be of the form
    start_index_tuple + a, where a is a multi-index with nonnegative components
    """
    def __init__(self, var_name, size_tuple, start_index_tuple=None,
                 reverse=False):
        if start_index_tuple is None:
            start_index_tuple = len(size_tuple) * (0, )
        cart = [()]
        assert len(size_tuple) == len(start_index_tuple)
        outer_indices = reversed(range(len(size_tuple)))
        for i in outer_indices:
            s_i = start_index_tuple[i]
            s = size_tuple[i]
            cart = [(j, ) + c for j in range(s_i, s_i + s) for c in cart]
        if reverse:
            cart.reverse()
        self.cart = cart
        self.cart2index = {v: k for k, v in enumerate(cart)}
        self.var_name = var_name
        self.names = [var_name + str(c) for c in cart]

    def __getitem__(self, i):
        return self.names[i]

    def __iter__(self):
        return iter(self.names)

    def __len__(self):
        return len(self.names)

    def register(self, start, context):
        def var_func(*indices):
            return Variable(self.cart2index[indices] + start)
        var_func.__name__ = self.var_name
        context[self.var_name] = var_func


class InOutBlock:
    def __init__(self, out_size, in_size, output='out', input='in',
                 in_start_index=0, out_start_index=0,
                 out_reverse=False, in_reverse=False):
        self.output = Block(var_name=output, start_index=out_start_index,
                            size=out_size, reverse=out_reverse)
        self.input = Block(var_name=input, start_index=in_start_index,
                           size=in_size, reverse=in_reverse)
        self.out_start_index = out_start_index

        self.in_start_index = in_start_index

    def __iter__(self):
        return chain(self.output, self.input)

    def __getitem__(self, i):
        if i < len(self.output):
            return self.output[i]
        return self.input[i - len(self.output)]

    def __len__(self):
        return len(self.output) + len(self.input)

    def register(self, start, context):
        self.output.register(start, context)
        self.input.register(start + len(self.output), context)
        self.out_vars = shift(context[self.output.var_name], self.
                              out_start_index)
        self.in_vars = shift(context[self.input.var_name], self.in_start_index)


class MultiBlock:
    def __init__(self, sizes=None, var_names=["v"],
                 start_indices=[], reverses=None):
        if reverses is None:
            reverses = []
        if sizes is None:
            sizes = []
        self.start_indices = start_indices + [0] * (len(var_names) -
                                                    len(start_indices))
        reverses += [False] * (len(var_names) - len(reverses))
        sizes += [1] * (len(var_names) - len(sizes))

        self.blocks = [Block(var_name=var_names[idx], size=sizes[idx],
                             start_index=self.start_indices[idx],
                             reverse=reverses[idx])
                       for idx in range(len(var_names))]

    def __iter__(self):
        return chain(*self.blocks)

    def __getitem__(self, i):
        return next(islice(chain(*self.blocks), i, i + 1))
        # sum([bl.names for bl in self.blocks])[i]

    def __len__(self):
        return sum(len(bl) for bl in self.blocks)

    def register(self, start, context):
        offset = 0
        for bl in self.blocks:
            bl.register(start + offset, context)
            offset += len(bl)

        self.vars = [shift(context[self.blocks[idx].var_name],
                           self.start_indices[idx])
                     for idx in range(len(self.blocks))]


class PrefixedDictProxy:
    """docstring for PrefixedDictProxy"""

    def __init__(self, wrapped, prefix):
        super().__init__()
        self.wrapped = wrapped
        self.prefix = prefix

    def __getitem__(self, k):
        try:
            return self.wrapped[self.prefix + k]
        except KeyError:
            print(self.prefix, k, list(self.wrapped))
            raise KeyError

    def __setitem__(self, k, v):
        self.wrapped[self.prefix + k] = v


class MacroBlock:
    def __init__(self, prefix):

        self.prefix = prefix
        self.blocks = []
        self.combinations = []
        self.connections = []

    def declare(self, blocks):
        self.blocks = blocks

    def connect(self, combinations):
        self.combinations = combinations

    def __iter__(self):
        return (self.prefix + "_" + n for n in chain(*self.blocks))

    def __getitem__(self, i):
        return self.prefix + "_" + next(islice(chain(*self.blocks), i, i + 1))

    def __len__(self):
        return sum(len(bl) for bl in self.blocks)

    def resolve(self, localname):
        return self.prefix + "_" + localname

    def register(self, start, context):
        context = PrefixedDictProxy(context, self.prefix + "_")
        offset = 0
        for bl in self.blocks:
            bl.register(start + offset, context)
            offset += len(bl)

        for ((con1, indices1), (con2, indices2)) in self.combinations:
            for idx in range(min(len(indices1), len(indices2))):
                self.connections += [context[con1](indices1[idx]) + context[
                    con2](indices2[idx])]

    def implement(self, equations):
        for bl in self.blocks:
            if hasattr(bl, "implement"):
                bl.implement(equations)

        equations += self.connections


class IfThen:
    def __init__(self, ifpart, thenpart, supposed_to_be_valid=True):
        self.ifpart = [Polynomial(p) for p in ifpart]
        self.thenpart = [Polynomial(p) for p in thenpart]
        self.supposedToBeValid = supposed_to_be_valid

    def __str__(self):
        return ("If(AND(" + ", ".join(f"{p} == 0" for p in self.ifpart) +
                ")), THEN " + ", ".join(f"{p} == 0" for p in self.thenpart))


def if_then(i, t, supposed_to_be_valid=True):
    return IfThen(i, t, supposed_to_be_valid)


def declare_ring(blocks, context=None):
    r"""
    Declare Ring is the preferred function to create a ring and declare a variable scheme,
    the number of variables is automatically determined, usually you pass globals() as context
    argument to store the ring and the variable mapping.

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori import *
        sage: declare_ring([Block("x",10),Block("y",5)],globals())
        Boolean PolynomialRing in x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, y0, y1, y2, y3, y4

    gives  a ring with x(0..9),y(0..4) and registers the ring as r, and the variable
    blocks x and y in the context dictionary globals(), which consists of the global
    variables of the python module
    """
    if context is None:
        context = sys.modules['__main__'].__dict__

    def canonicalize(blocks):
        for elt in blocks:
            if isinstance(elt, str):
                yield elt
            else:
                yield from elt

    blocks = list(blocks)
    n = 0

    for b in blocks:
        n = n + 1 if isinstance(b, str) else n + len(b)

    r = Ring(n, names=canonicalize(blocks))

    context["internalVariable"] = VariableFactory(r)
#  context["Monomial"] = MonomialFactory(r)
    context["r"] = r
    declare_block_scheme(blocks, context)
    return r


def declare_block_scheme(blocks, context):
    start = 0
    block_starts = []
    for b in blocks:
        if start:
            block_starts.append(start)
        if isinstance(b, str):
            context[b] = context["internalVariable"](start)
            start = start + 1
        else:
            b.register(start, context)
            start = start + len(b)
    context["block_start_hints"] = block_starts
    context["number_of_declared_vars"] = start


def main():
    r = Ring(1000)
    ablock = AlternatingBlock(["a", "b", "c"], 100)
    declare_block_scheme([ablock], globals())
    for i in range(10):
        print(r.variable(i))

    print(list(ablock))
    declare_block_scheme([Block(var_name="x", size=100),
                          HigherOrderBlock("y", (3, 4, 11, 2)),
                          AlternatingBlock(["a", "b", "c"], 100)],
                         globals())
    for i in range(10):
        print(x(i))
    print(y(0, 0, 0, 0))
    print(y(0, 0, 0, 1))
    print(y(0, 0, 1, 0))
    print(y(0, 0, 1, 1))
    print(a(0), a(1), a(2), b(0), b(1), c(0))
    declare_block_scheme([Block(var_name="x", size=100, reverse=True),
                          HigherOrderBlock("y", (3, 4, 11, 2), reverse=True),
                          AlternatingBlock(["a", "b", "c"], 100, reverse=True)],
                         globals())
    for i in range(10):
        print(x(i))
    print(y(0, 0, 0, 0))
    print(y(0, 0, 0, 1))
    print(y(0, 0, 1, 0))
    print(y(0, 0, 1, 1))
    print(a(0), a(1), a(2), b(0), b(1), c(0))
    declare_block_scheme(["a", "b", "c"], globals())
    print(a, b, c)


if __name__ == '__main__':
    main()



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from .pbori import BooleSet, FGLMStrategy
from .PyPolyBoRi import BoolePolynomialVector, Polynomial


def _fglm(I, from_ring, to_ring):
    r"""
    Unchecked variant of fglm
    """
    vec = BoolePolynomialVector(I)
    return FGLMStrategy(from_ring, to_ring, vec).main()


def fglm(I, from_ring, to_ring):
    r"""
    Convert *reduced* Groebner Basis in ``from_ring`` to a GroebnerBasis
    in ``to_ring``.

    It acts independent of the global ring, which is restored at the end of the
    computation.

    TESTS::

        sage: from sage.rings.polynomial.pbori import *
        sage: from sage.rings.polynomial.pbori.PyPolyBoRi import OrderCode
        sage: dp_asc = OrderCode.dp_asc
        sage: r=declare_ring(['x','y','z'],dict())
        sage: old_ring = r
        sage: new_ring = old_ring.clone(ordering=dp_asc)
        sage: (x,y,z) = [old_ring.variable(i) for i in range(3)]
        sage: ideal=[x+z, y+z]# lp Groebner basis
        sage: from sage.rings.polynomial.pbori.fglm import fglm
        sage: list(fglm(ideal, old_ring, new_ring))
        [y + x, z + x]
    """
    for poly in I:
        if poly.ring().id() != from_ring.id():
            raise ValueError("Ideal I must be from the first ring argument")
    return _fglm(I, from_ring, to_ring)


def vars_real_divisors(monomial, monomial_set):
    r"""
    Return all elements of ``monomial_set``, which result multiplied by a variable in monomial.

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori.pbori import *
        sage: from sage.rings.polynomial.pbori.PyPolyBoRi import OrderCode
        sage: dp_asc = OrderCode.dp_asc
        sage: from sage.rings.polynomial.pbori.PyPolyBoRi import Ring
        sage: r = Ring(1000)
        sage: x = r.variable
        sage: b = BooleSet([x(1)*x(2),x(2)])
        sage: from sage.rings.polynomial.pbori.fglm import vars_real_divisors
        sage: vars_real_divisors(x(1)*x(2)*x(3),b)
        {{x(1),x(2)}}
    """
    return BooleSet(Polynomial(monomial_set.divisors_of(monomial)).
                    graded_part(monomial.deg() - 1))


def m_k_plus_one(completed_elements, variables):
    r"""
    Calculate `m_{k+1}` from the FGLM algorithm.

    Calculate `m_{k+1}` from the FGLM algorithm as described in Wichmann [Wich1997]_.

    .. NOTE::

        It would be nice to be able to efficiently extract the smallest term of a polynomial.

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori.pbori import *
        sage: from sage.rings.polynomial.pbori.PyPolyBoRi import OrderCode
        sage: from sage.rings.polynomial.pbori.fglm import m_k_plus_one
        sage: dp_asc = OrderCode.dp_asc
        sage: from sage.rings.polynomial.pbori.PyPolyBoRi import Ring
        sage: r = Ring(1000)
        sage: x = r.variable
        sage: from sage.rings.polynomial.pbori.PyPolyBoRi import Monomial
        sage: s = BooleSet([x(1)*x(2),x(1),x(2),Monomial(r),x(3)])
        sage: variables = BooleSet([x(1),x(2),x(3)])
        sage: m_k_plus_one(s,variables)
        x(2)*x(3)
        sage: r2 = r.clone(ordering=dp_asc)
        sage: m_k_plus_one(r2(s).set(),r2(variables).set())
        x(1)*x(3)
    """
    return sorted(completed_elements.cartesian_product(variables).diff(
        completed_elements))[0]
